"""
Trafaret validators that doesn't part of the class.
Separated to own file to improve readability and reusability.
"""

import trafaret as t

# Representation of a single feature impact
# featureName - feature we are describing
# impactUnnormalized - drop in the metric value comparing to base
# impactNormalized - normalized value so that highest drop = 1.0
# redundantWith - one of 2 options
# - None if redundant detection wasn't run or
# - feature name that makes this one redundant
single_feature_impact_trafaret = t.Dict({
    t.Key('impactUnnormalized'): t.Float,
    t.Key('impactNormalized'): t.Float,
    t.Key('redundantWith'): t.Or(t.Null, t.String),
    t.Key('featureName'): t.String
}).ignore_extra('*')

# Feature impact data for regression, binary classification and aggregated score for multiclass
# ranRedundancyDetection - shows if redundancy detection was run for the model
# featureImpacts - list of impacts for each feature
feature_impact_trafaret = t.Dict({
    t.Key('count'): t.Int,
    t.Key('ranRedundancyDetection'): t.Bool,
    t.Key('rowCount'): t.Or(t.Int, t.Null),
    t.Key('shapBased'): t.Bool,
    t.Key('featureImpacts'): t.List(single_feature_impact_trafaret),
}).ignore_extra('*')

# Feature impact data for each class in multiclass model
# classFeatureImpacts - list of records for each class, with 2 keys
# - class - target class name
# - featureImpacts - list of impacts of each feature for the class
multiclass_feature_impact_trafaret = t.Dict({
    t.Key('ranRedundancyDetection'): t.Bool,
    t.Key('classFeatureImpacts'): t.List(t.Dict({
        t.Key('featureImpacts'): t.List(single_feature_impact_trafaret),
        t.Key('class'): t.String
    }).ignore_extra('*')),
}).ignore_extra('*')



import pytest
import responses
from mock import patch

from datarobot import ExternalMulticlassLiftChart

from datarobot.errors import ClientError
from datarobot.utils import from_api


class TestExternalMulticlassLiftChart(object):
    @pytest.fixture()
    def multiclass_liftchart_classbins_data(self, lift_chart_bins_data):
        return [
            {'targetClass': 'classA', 'bins': lift_chart_bins_data},
            {'targetClass': 'classB', 'bins': lift_chart_bins_data},
            {'targetClass': 'classC', 'bins': lift_chart_bins_data},
        ]

    @pytest.fixture()
    def one_class_liftchart(self, lift_chart_bins_data):
        return {
            'targetClass': 'classA',
            'bins': lift_chart_bins_data,
            'datasetId': 'dataset_id_1',
        }

    @staticmethod
    def create_response_data(multiclass_liftchart_classbins_data, count):
        return [
            {
                'classBins': multiclass_liftchart_classbins_data,
                'datasetId': 'dataset_id_{}'.format(i),
            }
            for i in range(count)
        ]

    @pytest.fixture
    def url(self, project_id, model_id):
        return 'https://host_name.com/projects/{}/models/{}/datasetMulticlassLiftCharts/'.format(
            project_id, model_id
        )

    def test_instantiation(self, one_class_liftchart):
        lc = ExternalMulticlassLiftChart.from_server_data(one_class_liftchart)
        assert lc.dataset_id == one_class_liftchart['datasetId']
        assert lc.bins == from_api(one_class_liftchart['bins'])
        assert lc.target_class == from_api(one_class_liftchart['targetClass'])

    def test_future_proof(self, one_class_liftchart):
        data_with_future_keys = dict(
            one_class_liftchart, new_key='some future lift data'
        )
        data_with_future_keys['bins'][0]['new_key'] = 'some future bin data'
        ExternalMulticlassLiftChart.from_server_data(data_with_future_keys)

    @responses.activate
    @patch('datarobot.models.external_dataset_scores_insights.external_multiclass_lift_chart.'
           'DEFAULT_BATCH_SIZE', 10)
    def test_list_lift_chart_all_results(
        self, multiclass_liftchart_classbins_data, url, project_id, model_id
    ):
        url_page1 = url + '?limit=10&offset=0'
        url_page2 = url + '?limit=10&offset=10'
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET,
            url_page1,
            json=resp_data_1,
            status=200,
            match_querystring=True,
        )
        responses.add(
            responses.GET,
            url_page2,
            json=resp_data_2,
            status=200,
            match_querystring=True,
        )

        lift_charts = ExternalMulticlassLiftChart.list(project_id, model_id, limit=0)
        n_classes = len(multiclass_liftchart_classbins_data)
        assert len(lift_charts) == 11 * n_classes
        for j in range(11):
            for k, class_bin in enumerate(multiclass_liftchart_classbins_data):
                i = j * n_classes + k
                assert isinstance(lift_charts[i], ExternalMulticlassLiftChart)
                assert lift_charts[i].dataset_id == 'dataset_id_{}'.format(j)
                assert lift_charts[i].bins == from_api(class_bin['bins'])
                assert lift_charts[i].target_class == class_bin['targetClass']

    @responses.activate
    def test_list_lift_chart_limited(
        self,
        multiclass_liftchart_classbins_data,
        lift_chart_bins_data,
        url,
        project_id,
        model_id,
    ):
        url_page = url + '?limit=3&offset=2'
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(
            responses.GET, url_page, json=resp_data, status=200, match_querystring=True
        )
        lift_charts = ExternalMulticlassLiftChart.list(
            project_id, model_id, limit=3, offset=2
        )
        n_classes = len(multiclass_liftchart_classbins_data)
        assert len(lift_charts) == 3 * n_classes
        for j in range(3):
            for k, class_bin in enumerate(multiclass_liftchart_classbins_data):
                i = j * n_classes + k
                assert isinstance(lift_charts[i], ExternalMulticlassLiftChart)
                assert lift_charts[i].dataset_id == 'dataset_id_{}'.format(j + 2)
                assert lift_charts[i].bins == from_api(class_bin['bins'])
                assert lift_charts[i].target_class == class_bin['targetClass']

    @responses.activate
    def test_get_lift_chart_by_dataset(
        self,
        multiclass_liftchart_classbins_data,
        lift_chart_bins_data,
        url,
        project_id,
        model_id,
    ):
        dataset_id = 'dataset_id_0'
        url_page = url + '?datasetId={}&limit=100&offset=0'.format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(
            responses.GET, url_page, json=resp_data, status=200, match_querystring=True
        )
        # chart for each class
        lift_charts = ExternalMulticlassLiftChart.list(project_id, model_id, dataset_id)
        assert len(lift_charts) == len(multiclass_liftchart_classbins_data)
        for chart in lift_charts:
            assert isinstance(chart, ExternalMulticlassLiftChart)
            assert chart.dataset_id == dataset_id
            assert chart.bins == from_api(lift_chart_bins_data)

    @responses.activate
    def test_get_lift_chart_by_target_class(
        self,
        multiclass_liftchart_classbins_data,
        lift_chart_bins_data,
        url,
        project_id,
        model_id,
    ):
        dataset_id = 'dataset_id_0'
        url_page = url + '?datasetId={}&limit=100&offset=0'.format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(
            responses.GET, url_page, json=resp_data, status=200, match_querystring=True
        )

        lift_chart = ExternalMulticlassLiftChart.get(
            project_id, model_id, dataset_id, target_class='classB'
        )
        assert isinstance(lift_chart, ExternalMulticlassLiftChart)
        assert lift_chart.dataset_id == dataset_id
        assert lift_chart.bins == from_api(lift_chart_bins_data)
        assert lift_chart.target_class == 'classB'

    @responses.activate
    def test_get_lift_chart_by_target_class_failure(
        self,
        multiclass_liftchart_classbins_data,
        lift_chart_bins_data,
        url,
        project_id,
        model_id,
    ):
        dataset_id = 'dataset_id_0'
        url_page = url + '?datasetId={}&limit=100&offset=0'.format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(
            responses.GET, url_page, json=resp_data, status=200, match_querystring=True
        )
        with pytest.raises(
            ClientError, match='Requested multiclass lift chart does not exist.'
        ):
            ExternalMulticlassLiftChart.get(
                project_id, model_id, dataset_id, target_class='wrong_class'
            )



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class FeatureAssociation(APIObject):
    """ Feature association statistics for a project.

    Attributes
    ----------
    type : str
        Either 'association' or 'correlation' the class of the pairwise stats
    metric : str
        the metric of either class of pairwise stats
        'spearman', 'pearson', etc for correlation, 'mutualInfo', 'cramersV'
        for association
    """
    _converter = t.Dict({
        t.Key('type'): t.Or(t.String, t.Null),
        t.Key('metric'): t.Or(t.String, t.Null),
        t.Key('featurelistId'): t.Or(t.String, t.Null)
    })

    def __init__(self, metric=None, assoc_type=None, featurelistId=None):
        self.metric = metric if metric else 'mutualInfo'
        self.assoc_type = type if type else 'association'
        self.featurelistId = featurelistId

    def __repr__(self):
        return encode_utf8_if_py2(u'FeatureAssociation({}/{}/{})'.format(
            self.assoc_type, self.metric, self.featurelistId))


class FeatureAssociationMatrixDetails(APIObject):
    """ Plotting details for a pair of passed features present in the feature
    association matrix

    Attributes
    ----------
    feature1 : str
        Feature name for the first feature of interest
    feature2 : str
        Feature name for the second feature of interest
    """
    _converter = t.Dict({
        t.Key('feature1'): t.String,
        t.Key('feature2'): t.String
    })

    def __init__(self, feature1=None, feature2=None):
        self.feature1 = feature1
        self.feature2 = feature2

    def __repr__(self):
        return encode_utf8_if_py2(u'FeatureAssociationMatrixDetails({}/{})'.format(
            self.feature1, self.feature2))


class FeatureAssociationFeaturelists(APIObject):
    """ Get project featurelists and see if they have association statistics
    """
    def __repr__(self):
        return encode_utf8_if_py2(u'FeatureAssocationFeaturelists')



import csv
import io
import json
import six
import tempfile

import pandas as pd
import pytest
import responses
import trafaret

from datarobot import TrainingPredictions
from datarobot.models.training_predictions import TrainingPredictionsRow


def _write_csv(lines):

    if six.PY2:

        buf = io.BytesIO()
        wr = csv.writer(buf)
        wr.writerows(lines)
        return buf.getvalue()

    buf = io.StringIO()
    wr = csv.writer(buf)
    wr.writerows(lines)
    return buf.getvalue().encode('utf-8')


@pytest.fixture
def prediction_id():
    return '59d92fe8962d7464ca8c6bd6'


@pytest.fixture
def predictions_with_shap_exp_csv():
    return """row_id,partition_id,prediction,class_1.0,class_0.0,explained_class,Explanation_1_feature_name,Explanation_1_feature_value,Explanation_1_strength,shap_remaining_total,shap_base_value
        0,TrainTestOverlap,0.0,0.311178247734139,0.6888217522658611,1.0,number_diagnoses,9,-0.07643818895488698,0.006562548084582246,0.39640626311300003
        1,TrainTestOverlap,0.0,0.2829861111111111,0.7170138888888888,1.0,number_diagnoses,6,0.06648347981438553,-0.013534302788801085,0.39640626311300003
        2,TrainTestOverlap,0.0,0.42699724517906334,0.5730027548209367,1.0,number_diagnoses,9,0.08556893050818416,-0.032619753482599656,0.39640626311300003
        3,TrainTestOverlap,0.0,0.32653061224489793,0.6734693877551021,1.0,number_diagnoses,3,0.05807900167179779,-0.0051298246462132885,0.39640626311300003
        4,TrainTestOverlap,0.0,0.06,0.94,1.0,number_inpatient,0,-0.056383896138088974,-0.0570362457984236,0.39640626311300003
        5,TrainTestOverlap,0.0,0.3774597495527728,0.6225402504472273,1.0,number_diagnoses,8,0.04847090471521205,-0.03250602234256627,0.39640626311300003
        6,TrainTestOverlap,0.0,0.13768115942028986,0.8623188405797102,1.0,admission_type_id,Elective,0.08996873371035077,0.040252262429918834,0.39640626311300003
        7,TrainTestOverlap,1.0,0.5300353356890459,0.4699646643109541,1.0,admission_type_id,Emergency,0.16263335056706338,-0.03241235442679378,0.39640626311300003
        8,TrainTestOverlap,1.0,0.6408668730650154,0.35913312693498456,1.0,number_inpatient,1,0.34801989552094514,-0.055934082238188985,0.39640626311300003
        9,TrainTestOverlap,0.0,0.2027972027972028,0.7972027972027972,1.0,number_inpatient,0,0.1146691220002886,-0.08407812406711129,0.39640626311300003
    """


@responses.activate
def test_get__no_requests_made(project_id, prediction_id):
    TrainingPredictions.get(project_id, prediction_id)
    assert len(responses.calls) == 0


@responses.activate
def test_list__ok(project_id, model_id, prediction_id):
    training_prediction_stub = [
        {
            'url': 'https://host_name.com/projects/{}/trainingPredictions/{}//'.format(
                project_id, prediction_id
            ),
            'id': prediction_id,
            'modelId': model_id,
            'dataSubset': 'all',
        },
        {
            'url': 'https://host_name.com/projects/{}/trainingPredictions/{}//'.format(
                project_id, prediction_id
            ),
            'id': prediction_id,
            'modelId': model_id,
            'dataSubset': 'all',
            'explanationAlgorithm': 'shap',
            'maxExplanations': 7,
            'shapWarnings': {
                'mismatchRowCount': 1,
                'maxNormalizedMismatch': 0.01,
            }
        }
    ]
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/trainingPredictions/'.format(
            project_id,
        ),
        status=200,
        body=json.dumps({'data': training_prediction_stub}),
    )

    training_predictions_list = TrainingPredictions.list(project_id)

    assert len(training_predictions_list) == 2
    training_prediction = training_predictions_list[0]
    assert training_prediction.project_id == project_id
    assert training_prediction.model_id == model_id
    assert training_prediction.prediction_id == prediction_id
    training_prediction_with_shap_exp = training_predictions_list[1]
    assert training_prediction_with_shap_exp.project_id == project_id
    assert training_prediction_with_shap_exp.model_id == model_id
    assert training_prediction_with_shap_exp.explanation_algorithm == 'shap'
    assert training_prediction_with_shap_exp.max_explanations == 7
    assert training_prediction_with_shap_exp.shap_warnings == {
        'mismatch_row_count': 1,
        'max_normalized_mismatch': 0.01,
    }


@responses.activate
def test_list__no_url_in_api_response__error(project_id, model_id, prediction_id):
    training_prediction_stub = {
        'url': '',
        'id': prediction_id,
        'modelId': model_id,
        'trainingPredictionSet': 'all',
    }
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/trainingPredictions/'.format(
            project_id,
        ),
        status=200,
        body=json.dumps({'data': [training_prediction_stub]}),
    )

    with pytest.raises(trafaret.DataError):
        TrainingPredictions.list(project_id)


@responses.activate
def test_iterate_rows__empty_predictions__ok(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response = {'data': [], 'next': '{}?offset=100&limit=100'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == []


@responses.activate
def test_iterate_rows__malfromed_predictions__data_error(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [{
        'row_id': 'wrong-id',
        'partition_id': 'holdout',
        'prediction': 0.42,
        'prediction_values': [],
    }]
    response = {'data': items, 'next': '{}?offset=100&limit=100'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    with pytest.raises(trafaret.DataError):
        obj = TrainingPredictions.get(project_id, prediction_id)
        it = obj.iterate_rows()
        next(it)


@responses.activate
def test_iterate_rows__extra_fields__ignored(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [{
        'row_id': 1,
        'partition_id': 2,
        'prediction': 0.42,
        'prediction_values': [{
            'label': 'test-label',
            'value': 0.42,
            'extra-prediction': 'you are going to win a lottery for sure',
        }]},
    ]
    response = {'data': items, 'next': '{}?offset=100&limit=100'.format(url), 'prev': None}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    assert list(obj.iterate_rows()) == [TrainingPredictionsRow(
        row_id=1,
        partition_id=2,
        prediction=0.42,
        prediction_values=[{
            'label': 'test-label',
            'value': 0.42,
        }],
        timestamp=None,
        forecast_point=None,
        forecast_distance=None,
        series_id=None,
        prediction_explanations=None,
        shap_metadata=None,
    )]


@responses.activate
def test_iterate_rows__all_rows_at_once__ok(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [
        {'row_id': 1, 'partition_id': 2, 'prediction': 0.42, 'prediction_values': []},
        {'row_id': 2, 'partition_id': 2, 'prediction': 0.22, 'prediction_values': []},
        {'row_id': 3, 'partition_id': 2, 'prediction': 0.82, 'prediction_values': []},
    ]
    response = {'data': items, 'next': None, 'prev': None}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    assert len(list(obj.iterate_rows(batch_size=3))) == 3


@responses.activate
def test_iterate_rows__one_prediction__ok(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [
        {'row_id': 1, 'partition_id': 2, 'prediction': 0.42, 'prediction_values': []},
    ]
    response = {'data': items, 'next': '{}?offset=100&limit=100'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == [
        TrainingPredictionsRow(
            row_id=1,
            partition_id=2,
            prediction=0.42,
            prediction_values=[],
            timestamp=None,
            forecast_point=None,
            forecast_distance=None,
            series_id=None,
            prediction_explanations=None,
            shap_metadata=None,
        ),
    ]
    assert len(responses.calls) == 1


@responses.activate
def test_iterate_rows__one_prediction_with_shap_exp(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [
        {
            'row_id': 1,
            'partition_id': 2,
            'prediction': 0.42,
            'prediction_values': [],
            'prediction_explanations': [
                {
                    'feature': "first_name",
                    'feature_value': "some_value",
                    'strength': -0.07,
                    'label': 1.0,
                },
                {
                    'feature': "second_name",
                    'feature_value': "0",
                    'strength': -0.05,
                    'label': 1.0,
                },
            ],
            'shap_metadata': {
                'shap_base_value': 0.39,
                'shap_remaining_total': 0.03,
                'warnings': {'mismatch_row_count': 1, 'max_normalized_mismatch': 0.01},
            },

        },
    ]
    response = {'data': items, 'next': '{}?offset=100&limit=100'.format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type='application/json'
    )

    training_predictions = TrainingPredictions.get(project_id, prediction_id)

    data = list(training_predictions.iterate_rows())
    assert len(data) == 1
    training_prediction = data[0]
    assert training_prediction.row_id == 1
    assert training_prediction.partition_id == 2
    assert training_prediction.prediction == 0.42
    assert training_prediction.prediction_values == []
    assert training_prediction.timestamp is None
    assert training_prediction.forecast_point is None
    assert training_prediction.forecast_distance is None
    assert training_prediction.series_id is None
    assert training_prediction.series_id is None
    assert training_prediction.prediction_explanations == [
        {
            'feature': "first_name",
            'feature_value': "some_value",
            'strength': -0.07,
            'label': 1.0,
        },
        {
            'feature': "second_name",
            'feature_value': "0",
            'strength': -0.05,
            'label': 1.0,
        }
    ]
    assert training_prediction.shap_metadata == {
        'shap_base_value': 0.39,
        'shap_remaining_total': 0.03,
        'warnings': {'mismatch_row_count': 1, 'max_normalized_mismatch': 0.01}
    }
    assert len(responses.calls) == 1


@responses.activate
def test_iterate_rows__two_predictions_with_limit_equals_one__three_requests_were_made(
        project_id, prediction_id,
):
    items = [
        {'row_id': 1, 'partition_id': 2, 'prediction': 0.42, 'prediction_values': []},
        {'row_id': 2, 'partition_id': 2, 'prediction': 0.22, 'prediction_values': []},
    ]
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    body = {'data': items[:1], 'next': '{}?offset=1&limit=1'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    it = obj.iterate_rows(batch_size=1)

    assert next(it, None) == TrainingPredictionsRow(
        row_id=1,
        partition_id=2,
        prediction=0.42,
        prediction_values=[],
        timestamp=None,
        forecast_point=None,
        forecast_distance=None,
        series_id=None,
        prediction_explanations=None,
        shap_metadata=None,
    )
    assert len(responses.calls) == 1

    body2 = {'data': items[1:], 'next': '{}?offset=2&limit=1'.format(url)}
    responses.reset()
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body2),
        status=200,
        content_type='application/json',
        match_querystring=False,
    )

    assert next(it, None) == TrainingPredictionsRow(
        row_id=2,
        partition_id=2,
        prediction=0.22,
        prediction_values=[],
        timestamp=None,
        forecast_point=None,
        forecast_distance=None,
        series_id=None,
        prediction_explanations=None,
        shap_metadata=None,
    )
    assert len(responses.calls) == 1

    body = {'data': [], 'next': '{}?offset=3&limit=1'.format(url)}
    responses.reset()
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body),
        status=200,
        content_type='application/json',
    )

    assert next(it, None) is None
    assert len(responses.calls) == 1


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [
            {'row_id': 1, 'partition_id': 2, 'prediction': 0.42, 'prediction_values': []},
            {'row_id': 2, 'partition_id': 2, 'prediction': 0.22, 'prediction_values': []},
        ], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            ['row_id', 'partition_id', 'prediction'],
            [1, 2, 0.42],
            [2, 2, 0.22],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_get_all_as_dataframe__regression_two_predictions__ok(
        project_id, prediction_id,
        project_url, project_with_target_json,
        serializer, response,
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response['body'] = response['body'](url)
    responses.add(
        responses.GET,
        url,
        status=200,
        **response
    )
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 3)
    assert list(data_frame.columns) == [
        'row_id',
        'partition_id',
        'prediction',
    ]


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [
            {
                'row_id': 1,
                'partition_id': '1960-01-01T00:00:00.000000Z',
                'prediction': 0.42,
                'prediction_values': [],
                'timestamp': '1960-01-01T00:00:00.000000Z',
                'forecast_point': '1959-12-01T00:00:00.000000Z',
                'forecast_distance': 1,
                'series_id': "1",
            },
            {
                'row_id': 2,
                'partition_id': '1960-01-01T00:00:00.000000Z',
                'prediction': 0.22,
                'prediction_values': [],
                'timestamp': '1959-11-01T00:00:00.000000Z',
                'forecast_point': '1959-12-01T00:00:00.000000Z',
                'forecast_distance': 2,
                'series_id': "1",
            },
        ], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            [
                'row_id', 'partition_id', 'prediction', 'timestamp',
                'forecast_point', 'forecast_distance', 'series_id'
            ],
            [
                1,
                '1960-01-01T00:00:00.000000Z',
                0.42,
                '1960-01-01T00:00:00.000000Z',
                '1959-12-01T00:00:00.000000Z',
                1,
                "1",
            ],
            [
                2,
                '1960-01-01T00:00:00.000000Z',
                0.22,
                '1959-11-01T00:00:00.000000Z',
                '1959-12-01T00:00:00.000000Z',
                2,
                "1",
            ],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv'),
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [
            {
                'row_id': 1,
                'partition_id': '1960-01-01T00:00:00.000000Z',
                'prediction': 0.42,
                'prediction_values': [],
                'timestamp': '1960-01-01T00:00:00.000000Z',
                'forecast_point': '1959-12-01T00:00:00.000000Z',
                'forecast_distance': 1,
                'series_id': 1,
            },
            {
                'row_id': 2,
                'partition_id': '1960-01-01T00:00:00.000000Z',
                'prediction': 0.22,
                'prediction_values': [],
                'timestamp': '1959-11-01T00:00:00.000000Z',
                'forecast_point': '1959-12-01T00:00:00.000000Z',
                'forecast_distance': 2,
                'series_id': 1,
            },
        ], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            [
                'row_id', 'partition_id', 'prediction', 'timestamp',
                'forecast_point', 'forecast_distance', 'series_id'
            ],
            [
                1,
                '1960-01-01T00:00:00.000000Z',
                0.42,
                '1960-01-01T00:00:00.000000Z',
                '1959-12-01T00:00:00.000000Z',
                1,
                1,
            ],
            [
                2,
                '1960-01-01T00:00:00.000000Z',
                0.22,
                '1959-11-01T00:00:00.000000Z',
                '1959-12-01T00:00:00.000000Z',
                2,
                1,
            ],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_get_all_as_dataframe__timeseries_predictions__ok(
        project_id, prediction_id,
        project_url, project_with_target_json,
        serializer, response,
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response['body'] = response['body'](url)
    responses.add(
        responses.GET,
        url,
        status=200,
        **response
    )
    project_data = json.loads(project_with_target_json)
    project_data['partition']['use_time_series'] = True
    responses.add(
        responses.GET,
        project_url,
        body=json.dumps(project_data),
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 7)
    assert set(data_frame.columns) == {
        'row_id', 'partition_id', 'prediction', 'timestamp',
        'forecast_point', 'forecast_distance', 'series_id'
    }

    assert len(data_frame) == 2
    assert None not in data_frame.timestamp.unique()
    assert None not in data_frame.forecast_point.unique()
    assert all(isinstance(value, int) for value in data_frame.forecast_distance)


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [
            {'row_id': 1, 'partition_id': 2, 'prediction': 0.42, 'prediction_values': []},
            {'row_id': 2, 'partition_id': 2, 'prediction': 0.22, 'prediction_values': []},
        ], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            ['row_id', 'partition_id', 'prediction'],
            [1, 2, 0.42],
            [2, 2, 0.22],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_download_to_csv__two_regression_predictions__ok(
        project_id, prediction_id,
        project_url, project_with_target_json,
        serializer, response,
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response['body'] = response['body'](url)
    responses.add(
        responses.GET,
        url,
        status=200,
        **response
    )
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    with tempfile.NamedTemporaryFile(mode='r') as a_file:
        obj.download_to_csv(a_file.name, serializer=serializer)
        data_frame = pd.read_csv(a_file.name)

    assert data_frame.shape == (2, 3)
    assert set(data_frame.columns) == {'row_id', 'partition_id', 'prediction'}


@responses.activate
def test_iterate_rows__one_classification_prediction__ok(project_id, prediction_id):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    items = [
        {
            'row_id': 1,
            'partition_id': 2,
            'prediction': 0.42,
            'prediction_values': [
                {'label': 'good', 'value': 0.8},
                {'label': 'bad', 'value': 0.2},
            ],
        },
    ]
    response = {'data': items, 'next': '{}?offset=100&limit=100'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(response),
        status=200,
        content_type='application/json',
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == [
        TrainingPredictionsRow(
            row_id=1,
            partition_id=2,
            prediction=0.42,
            prediction_values=[
                {'label': 'good', 'value': 0.8},
                {'label': 'bad', 'value': 0.2},
            ],
            timestamp=None,
            forecast_point=None,
            forecast_distance=None,
            series_id=None,
            prediction_explanations=None,
            shap_metadata=None,
        ),
    ]
    assert len(responses.calls) == 1


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [{
            'row_id': 1,
            'partition_id': 2,
            'prediction': 0.42,
            'prediction_values': [
                {'label': 'good', 'value': 0.8},
                {'label': 'bad', 'value': 0.2},
            ],
        }], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            ['row_id', 'partition_id', 'prediction', 'foo_good', 'foo_bad'],
            [1, 2, 0.42, 0.8, 0.2],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_get_all_as_dataframe__one_classification_prediction__ok(
        project_id, prediction_id,
        project_url, project_with_target_json,
        serializer, response,
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response['body'] = response['body'](url)
    responses.add(
        responses.GET,
        url,
        status=200,
        **response
    )

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(
        class_prefix='foo_',
        serializer=serializer,
    )

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (1, 5)
    assert set(data_frame.columns) == {
        'row_id', 'partition_id', 'prediction', 'foo_good', 'foo_bad'
    }


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [{
            'row_id': 1,
            'partition_id': 2,
            'prediction': 0.42,
            'prediction_values': [
                {'label': 'good', 'value': 0.3},
                {'label': 'bad', 'value': 0.2},
                {'label': 'ugly', 'value': 0.2},
                {'label': 'none', 'value': 0.0},
            ],
        }], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            [
                'row_id', 'partition_id', 'prediction', 'class_good', 'class_bad',
                'class_ugly', 'class_none',
            ],
            [1, 2, 0.42, 0.3, 0.2, 0.2, 0.0],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_download_to_csv__multiclass_project__prediction__ok(
        project_id, prediction_id,
        project_url, project_with_target_json,
        serializer, response,
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id,
        prediction_id,
    )
    response['body'] = response['body'](url)
    responses.add(
        responses.GET,
        url,
        status=200,
        **response
    )

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    with tempfile.NamedTemporaryFile(mode='r') as a_file:
        obj.download_to_csv(a_file.name, serializer=serializer)
        data_frame = pd.read_csv(a_file.name)

    assert data_frame.shape == (1, 7)
    assert set(data_frame.columns) == {
        'row_id', 'partition_id', 'prediction', 'class_good', 'class_bad',
        'class_ugly', 'class_none',
    }


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [{
            'row_id': 0,
            'partition_id': "0.0",
            'prediction': 0.0,
            'prediction_threshold': 0.5,
            'prediction_values': [
                {'label': 1.0, 'value': 0.311178247734139},
                {'label': 0.0, 'value': 0.6888217522658611},
            ],
            'prediction_explanations': [
                {
                    'featureValue': '5',
                    'strength': -0.07643818895488697,
                    'feature': 'number_diagnoses',
                    'label': 1.0,
                },
            ],
            'shap_metadata': {
                "shap_base_value": 0.39640626311302185,
                "shap_remaining_total": 0.06644934718939229,
                "warnings": None,
            },
        }], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            [
                'row_id', 'partition_id', 'prediction', 'class_1.0', 'class_0.0', 'explained_class',
                'Explanation_1_feature_name', 'Explanation_1_feature_value',
                'Explanation_1_strength', 'shap_remaining_total', 'shap_base_value'
            ],
            [
                0, 0.0, 0.0, 0.311178247734139, 0.6888217522658611, 1.0, 'number_diagnoses', 5,
                -0.07643818895488697, 'number_inpatient', 0, -0.05988679910481005,
                0.06644934718939229, 0.396406263113
            ]
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_get_all_as_dataframe_one_classification_prediction_with_explanations(
        project_id, prediction_id, project_url, project_with_target_json, serializer, response
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id, prediction_id
    )
    response['body'] = response['body'](url)
    responses.add(responses.GET, url, status=200, **response)

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (1, 11)
    assert set(data_frame.columns) == {
        'row_id', 'partition_id', 'prediction', 'class_1.0', 'class_0.0', 'explained_class',
        'shap_remaining_total', 'shap_base_value',
        'Explanation_1_feature_name', 'Explanation_1_feature_value', 'Explanation_1_strength',
    }


@responses.activate
@pytest.mark.parametrize(['serializer', 'response'], [
    pytest.param('json', {
        'body': lambda url: json.dumps({'data': [
            {
                'row_id': 0,
                'partition_id': '0.0',
                'prediction': 71700.0,
                'prediction_values': [
                    {'value': 71700.0, 'label': 'SalePrice'}
                ],
                'shap_metadata': {
                    'shap_base_value': 31108.28125,
                    'shap_remaining_total': -4045.506156357904,
                    'warnings': None,
                },
                'prediction_explanations': [
                    {
                        'feature_value': '172632',
                        'strength': 9021.030323424387,
                        'feature': 'MachineID',
                        'label': 'SalePrice',
                    },
                    {
                        'feature_value': '3855',
                        'strength': -4883.805345350174,
                        'feature': 'ModelID',
                        'label': 'SalePrice',
                    }
                ],
            },
            {
                'row_id': 1,
                'partition_id': '0.0',
                'prediction': 12275.0,
                'prediction_values': [
                    {'value': 12275.0, 'label': 'SalePrice'}
                ],
                'shap_metadata': {
                    'shap_base_value': 31108.28125,
                    'shap_remaining_total': -3350.9689739794676,
                    'warnings': None,
                },
                'prediction_explanations': [
                    {
                        'feature_value': '1986',
                        'strength': -6307.912477762799,
                        'feature': 'YearMade',
                        'label': 'SalePrice',
                    },
                    {
                        'feature_value': '1190700',
                        'strength': -3257.09113279142,
                        'feature': 'MachineID',
                        'label': 'SalePrice',
                    }
                ],
            },
        ], 'next': '{}?limit=20'.format(url)}),
        'content_type': 'application/json',
    }, id='json'),
    pytest.param('csv', {
        'body': lambda url: _write_csv([
            [
                'row_id', 'partition_id', 'prediction', 'Explanation_1_feature_name',
                'Explanation_1_feature_value', 'Explanation_1_strength',
                'Explanation_2_feature_name', 'Explanation_2_feature_value',
                'Explanation_2_strength', 'shap_remaining_total', 'shap_base_value'
            ],
            [
                0, 0.0, 71700.0, 'MachineID', '172632', 9021.030323424387, 'ModelID', '3855',
                -4883.805345350174, -4045.506156357904, 31108.28125
            ],
            [
                1, 0.0, 12275.0, 'YearMade', '1986', -6307.912477762799, 'MachineID', '1190700',
                -3257.09113279142, -3350.9689739794676, 31108.28125
            ],
        ]),
        'content_type': 'text/csv',
        'stream': True,
    }, id='csv')
])
def test_get_all_as_dataframe_regression_predictions_with_explanations(
        project_id, prediction_id, project_url, project_with_target_json, serializer, response
):
    url = 'https://host_name.com/projects/{}/trainingPredictions/{}/'.format(
        project_id, prediction_id
    )
    response['body'] = response['body'](url)
    responses.add(responses.GET, url, status=200, **response)
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json',
    )

    predictions = TrainingPredictions.get(project_id, prediction_id)
    data_frame = predictions.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 11)
    assert list(data_frame.columns) == [
        'row_id', 'partition_id', 'prediction', 'Explanation_1_feature_name',
        'Explanation_1_feature_value', 'Explanation_1_strength', 'Explanation_2_feature_name',
        'Explanation_2_feature_value', 'Explanation_2_strength', 'shap_remaining_total',
        'shap_base_value'
    ]



import json
from six.moves.urllib.parse import urljoin

import random
import pytest


@pytest.fixture
def visualai_url(project_url):
    return urljoin(project_url, "images/")


@pytest.fixture
def visualai_image_url(visualai_url, visualai_image_id):
    return urljoin(visualai_url, visualai_image_id + "/")


@pytest.fixture
def visualai_image_id():
    return "5e7e562528513130ab237875"


@pytest.fixture
def visualai_embeddings_url(model_url):
    return urljoin(model_url, "imageEmbeddings/")


@pytest.fixture
def visualai_activationmaps_url(model_url):
    return urljoin(model_url, "imageActivationMaps/")


@pytest.fixture(params=["aim", "eda", "eda2", "empty", "modeling"])
def visualai_project(request, project_with_target_json):
    ret = json.loads(project_with_target_json)
    ret["stage"] = request.param
    return ret


@pytest.fixture
def visualai_image():
    return {"width": 256, "height": 256, "imageId": "5e7e562528513130ab237875"}


@pytest.fixture
def visualai_image_file():
    return bytes((random.randint(0, 255) for i in range(32)))


@pytest.fixture
def visualai_sample(visualai_project):
    if visualai_project["stage"] in ["eda2", "modeling"]:
        return {
            "next": None,
            "data": [
                {
                    "imageId": "5e7e562528513130ab237875",
                    "height": 256,
                    "width": 256,
                    "targetBinRowCount": 1000,
                    "targetBinStart": None,
                    "targetBinEnd": None,
                    "targetValue": "fake",
                },
                {
                    "imageId": "5e7e562528513130ab237874",
                    "height": 256,
                    "width": 256,
                    "targetBinRowCount": 1000,
                    "targetBinStart": None,
                    "targetBinEnd": None,
                    "targetValue": "real",
                },
            ],
            "previous": None,
        }

    else:
        return {
            "next": None,
            "data": [
                {"imageId": "5e7e562528513130ab237875", "height": 256, "width": 256},
                {"imageId": "5e7e562528513130ab237874", "height": 256, "width": 256},
            ],
            "previous": None,
        }


@pytest.fixture
def visualai_duplicate():
    return {
        "count": 0,
        "next": None,
        "data": [
            {"imageId": "5e7e562528513130ab237875", "rowCount": 3},
            {"imageId": "5e7e562528513130ab237874", "rowCount": 7},
        ],
        "previous": None,
    }


@pytest.fixture
def visualai_embeddings():
    return {
        "targetValues": ["fake", "real"],
        "targetBins": None,
        "embeddings": [
            {
                "imageId": "5e7e562b28513130ab23792e",
                "positionX": 0.5935041904449463,
                "positionY": 0.6990952491760254,
                "actualTargetValue": "fake",
            },
            {
                "imageId": "5e7e563528513130ab237a6d",
                "positionX": 0.708981454372406,
                "positionY": 0.6228484511375427,
                "actualTargetValue": "fake",
            },
        ],
    }


@pytest.fixture
def visualai_activationmaps():
    return {
        "activationMapWidth": 56,
        "activationMaps": [
            {
                "imageId": "5eaafb57fcdb565e4d778f29",
                "overlayImageId": "5e7e563528513130ab237a6d",
                "predictedTargetValue": "fake",
                "featureName": "image",
                "actualTargetValue": "fake",
                "imageHeight": 256,
                "imageWidth": 256,
                "activationValues": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            },
            {
                "imageId": "5eaafb58fcdb565e2f7790d0",
                "predictedTargetValue": "real",
                "featureName": "image",
                "actualTargetValue": "real",
                "imageHeight": 256,
                "imageWidth": 256,
                "activationValues": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            },
        ],
        "activationMapHeight": 56,
        "targetBins": None,
        "targetValues": ["fake", "real"],
    }



import json
from six.moves.urllib.parse import urljoin
import responses

from datarobot.models.visualai import Image, SampleImage, DuplicateImage


@responses.activate
def test_get_image(visualai_image_url, visualai_image_id, visualai_image, project_id):
    responses.add(
        responses.GET,
        visualai_image_url,
        status=200,
        content_type='application/json',
        body=json.dumps(visualai_image),
    )
    image = Image.get(project_id, visualai_image_id)
    assert isinstance(image, Image)
    assert image.id == visualai_image['imageId']
    assert image.project_id == project_id
    assert image.width == visualai_image['width']
    assert image.height == visualai_image['height']


@responses.activate
def test_get_image_bytes(
    visualai_image_url, visualai_image_id, visualai_image_file, project_id
):
    responses.add(
        responses.GET,
        urljoin(visualai_image_url, 'file/'),
        status=200,
        content_type='image/jpeg',
        body=visualai_image_file,
    )
    image = Image(
        project_id=project_id, image_id=visualai_image_id, width=256, height=256
    )
    assert image.image_type == 'image/jpeg'
    assert image.image_bytes == visualai_image_file


@responses.activate
def test_get_sample_images(visualai_project, visualai_sample, project_url, project_id):
    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type='application/json',
        body=json.dumps(visualai_project),
    )
    stage_url = {
        'aim': 'imageSamples/?featureName=image',
        'eda': 'imageSamples/?featureName=image',
        'eda2': 'images/?featureName=image',
        'empty': 'imageSamples/?featureName=image',
        'modeling': 'images/?featureName=image',
    }[
        visualai_project['stage']
    ]
    responses.add(
        responses.GET,
        urljoin(project_url, stage_url),
        status=200,
        content_type='application/json',
        body=json.dumps(visualai_sample),
    )
    img0, img1 = SampleImage.list(project_id, 'image')
    assert isinstance(img0, SampleImage)
    assert img0.image.id == visualai_sample.get('data')[0].get('imageId')
    assert img0.project_id == project_id
    assert img0.target_value == visualai_sample.get('data')[0].get('targetValue')
    assert isinstance(img1, SampleImage)
    assert img1.image.id == visualai_sample.get('data')[1].get('imageId')
    assert img1.project_id == project_id
    assert img1.target_value == visualai_sample.get('data')[1].get('targetValue')


@responses.activate
def test_get_duplicate_images(
    visualai_project, visualai_duplicate, project_url, project_id
):
    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type='application/json',
        body=json.dumps(visualai_project),
    )
    responses.add(
        responses.GET,
        urljoin(project_url, 'duplicateImages/image/'),
        status=200,
        content_type='application/json',
        body=json.dumps(visualai_duplicate),
    )
    img0, img1 = DuplicateImage.list(project_id, 'image')
    assert isinstance(img0, DuplicateImage)
    assert img0.image.id == visualai_duplicate.get('data')[0].get('imageId')
    assert img0.project_id == project_id
    assert img0.count == visualai_duplicate.get('data')[0].get('rowCount')
    assert isinstance(img1, DuplicateImage)
    assert img1.image.id == visualai_duplicate.get('data')[1].get('imageId')
    assert img1.project_id == project_id
    assert img1.count == visualai_duplicate.get('data')[1].get('rowCount')



import pytest
import responses
from mock import patch

from datarobot import ExternalScores
from datarobot.errors import ClientError


class TestExternalScores(object):

    @pytest.fixture()
    def scores_data(self, project_id, model_id):
        data = {'projectId': project_id,
                'modelId': model_id,
                'scores': [{'label': 'AUC', 'value': 1.0},
                           {'label': 'RMSE', 'value': 0.0606},
                           {'label': 'Kolmogorov-Smirnov', 'value': 1.0},
                           {'label': 'LogLoss', 'value': 0.01383},
                           {'label': 'FVE Binomial', 'value': -13845778590722.23},
                           {'label': 'Max MCC', 'value': 1.0}]
                }

        return data

    @staticmethod
    def create_response_data(count, scores, model_id=None, actual_value_column=None):
        data = []
        for i in range(count):
            item = dict(scores, datasetId='dataset_id_{}'.format(i),
                        actualValueColumn=actual_value_column)
            if model_id:
                item['modelId'] = model_id
            data.append(item)
        return data

    @pytest.fixture
    def url(self, project_id):
        return 'https://host_name.com/projects/{}/externalScores/'.format(project_id)

    def test_instantiation(self, scores_data):
        scores = self.create_response_data(1, scores_data, actual_value_column='label')[0]
        scores_obj = ExternalScores.from_server_data(scores)

        assert scores_obj.dataset_id == 'dataset_id_0'
        assert scores_obj.project_id == scores_data['projectId']
        assert scores_obj.model_id == scores_data['modelId']
        assert scores_obj.actual_value_column == 'label'
        assert scores_obj.scores == scores_data['scores']

    def test_future_proof(self, scores_data):
        scores = self.create_response_data(1, scores_data)[0]
        data_with_future_keys = dict(scores, new_key='some future lift data')
        ExternalScores.from_server_data(data_with_future_keys)

    @responses.activate
    def test_scores_insights_create(self, project_id, project_url, base_job_server_data):
        scores_base_url = '{}externalScores/'.format(project_url)
        job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])

        responses.add(responses.POST, scores_base_url,
                      body='', status=202, content_type='application/json',
                      adding_headers={'Location': job_url})
        responses.add(responses.GET, job_url, json=base_job_server_data, status=303,
                      adding_headers={'Location': job_url})
        insights_job = ExternalScores.create(project_id=project_id,
                                             model_id='578e59a41ced2e5a9eb18960',
                                             dataset_id='581c6256100d2b60586980d9')
        assert str(insights_job.id) == base_job_server_data['id']

    @responses.activate
    def test_scores_insights_create_empty_target_unsupervised(self,  project_id, project_url,
                                                              base_job_server_data):
        scores_base_url = '{}externalScores/'.format(project_url)
        job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])

        responses.add(responses.POST, scores_base_url,
                      body='', status=202, content_type='application/json',
                      adding_headers={'Location': job_url})
        exception_text = (
                    'Cannot compute scores and insights on this dataset because target column '
                    'is missing'
                )
        responses.add(responses.GET, job_url, json={'message': exception_text}, status=422,
                      adding_headers={'Location': job_url})
        with pytest.raises(ClientError, message=exception_text):
            ExternalScores.create(project_id=project_id,
                                  model_id='578e59a41ced2e5a9eb18960',
                                  dataset_id='581c6256100d2b60586980d9')

    @responses.activate
    @patch('datarobot.models.external_dataset_scores_insights.external_scores.DEFAULT_BATCH_SIZE',
           10)
    def test_external_scores_all_results(self, scores_data, url, project_id, model_id):
        url_page1 = url + '?modelId={}&limit=10&offset=0'.format(model_id)
        url_page2 = url + '?modelId={}&limit=10&offset=10'.format(model_id)
        all_data = self.create_response_data(11, scores_data)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(responses.GET, url_page1, json=resp_data_1, status=200,
                      match_querystring=True)
        responses.add(responses.GET, url_page2, json=resp_data_2, status=200,
                      match_querystring=True)

        scores_list = ExternalScores.list(project_id, model_id, limit=0)
        assert len(scores_list) == 11
        for i, scores in enumerate(scores_list):
            assert isinstance(scores, ExternalScores)
            assert scores.project_id == project_id
            assert scores.model_id == model_id
            assert scores.scores == scores_data['scores']

    @responses.activate
    @patch('datarobot.models.external_dataset_scores_insights.external_scores.DEFAULT_BATCH_SIZE',
           1)
    def test_external_scores_all_results_with_offset(self, scores_data, url, project_id):
        url_page1 = url + '?datasetId={}&limit=1&offset=1'.format('dataset_id_0')
        url_page2 = url + '?datasetId={}&limit=1&offset=2'.format('dataset_id_0')
        all_data = self.create_response_data(1, scores_data, model_id='model_id_0')
        all_data.extend(self.create_response_data(1, scores_data, model_id='model_id_1'))
        all_data.extend(self.create_response_data(1, scores_data, model_id='model_id_2'))
        resp_data_1 = dict(count=1, previous=None, next=url_page2, data=[all_data[1]])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=[all_data[2]])
        responses.add(responses.GET, url_page1, json=resp_data_1, status=200,
                      match_querystring=True)
        responses.add(responses.GET, url_page2, json=resp_data_2, status=200,
                      match_querystring=True)

        scores_list = ExternalScores.list(project_id, dataset_id='dataset_id_0', limit=0, offset=1)
        assert len(scores_list) == 2
        assert scores_list[0].dataset_id == 'dataset_id_0'
        assert scores_list[0].model_id == 'model_id_1'
        assert scores_list[1].dataset_id == 'dataset_id_0'
        assert scores_list[1].model_id == 'model_id_2'

    @responses.activate
    def test_external_scores_limited(self, scores_data, url, project_id, model_id):
        url_page = url + '?limit=3&offset=2'
        all_data = self.create_response_data(5, scores_data)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        scores_list = ExternalScores.list(project_id, limit=3, offset=2)
        assert len(scores_list) == 3
        for i, scores in enumerate(scores_list):
            assert isinstance(scores, ExternalScores)
            assert scores.dataset_id == 'dataset_id_{}'.format(i + 2)

    @responses.activate
    def test_get_scores_success(self, scores_data, url, project_id, model_id):
        dataset_id = 'dataset_id_0'
        url_page = url + '?modelId={}&datasetId={}&limit=100&offset=0'.format(model_id, dataset_id)
        all_data = self.create_response_data(2, scores_data, model_id=model_id)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        scores = ExternalScores.get(project_id, model_id, dataset_id)
        assert isinstance(scores, ExternalScores)
        assert scores.dataset_id == dataset_id
        assert scores.model_id == model_id

    @responses.activate
    def test_get_scores_failure(self, url, project_id, model_id):
        dataset_id = 'dataset_id_0'
        url_page = url + '?modelId={}&datasetId={}&limit=100&offset=0'.format(model_id, dataset_id)
        resp_data = dict(count=0, previous=None, next=None, data=[])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        with pytest.raises(ValueError, match='dataset_id must be specified'):
            ExternalScores.get(project_id, model_id, None)
        with pytest.raises(ValueError, match='model_id must be specified'):
            ExternalScores.get(project_id, None, None)
        with pytest.raises(ClientError, match='Requested scores do not exist.'):
            ExternalScores.get(project_id, model_id, dataset_id)



import json

import pytest
import responses
from datarobot.errors import ClientError

from datarobot.utils import from_api

from datarobot import Model

from datarobot.models.roc_curve import RocCurve


@pytest.fixture
def roc_curve_validation_data(model_id, roc_curve_data):
    return {
        'source': 'validation',
        'sourceModelId': model_id,
        'rocPoints': roc_curve_data['rocPoints'],
        'negativeClassPredictions': roc_curve_data['negativeClassPredictions'],
        'positiveClassPredictions': roc_curve_data['positiveClassPredictions']
    }


@pytest.fixture
def roc_curve_parent_model_validation_data(parent_model_id, roc_curve_data):
    return {
        'source': 'validation',
        'sourceModelId': parent_model_id,
        'rocPoints': roc_curve_data['rocPoints'],
        'negativeClassPredictions': roc_curve_data['negativeClassPredictions'],
        'positiveClassPredictions': roc_curve_data['positiveClassPredictions']
    }


@pytest.fixture
def roc_curve_parent_model_holdout_data(parent_model_id, roc_curve_data):
    return {
        'source': 'holdout',
        'sourceModelId': parent_model_id,
        'rocPoints': roc_curve_data['rocPoints'],
        'negativeClassPredictions': roc_curve_data['negativeClassPredictions'],
        'positiveClassPredictions': roc_curve_data['positiveClassPredictions']
    }


def test_instantiation(roc_curve_validation_data):
    roc = RocCurve.from_server_data(roc_curve_validation_data)

    assert roc.source == roc_curve_validation_data['source']
    assert roc.negative_class_predictions == roc_curve_validation_data['negativeClassPredictions']
    assert roc.positive_class_predictions == roc_curve_validation_data['positiveClassPredictions']
    assert roc.roc_points == from_api(roc_curve_validation_data['rocPoints'])
    assert roc.source_model_id == roc_curve_validation_data['sourceModelId']


def test_future_proof(roc_curve_validation_data):
    data_with_future_keys = dict(roc_curve_validation_data, new_key='some future roc data')
    data_with_future_keys['rocPoints'][0]['new_key'] = 'some future bin data'
    RocCurve.from_server_data(data_with_future_keys)


@pytest.fixture
def roc_curve_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/rocCurve/validation/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def roc_curve_parent_model_validation_data_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/rocCurve/validation/'.format(
        project_id,
        parent_model_id
    )


@pytest.fixture
def roc_curve_list_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/rocCurve/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def roc_curve_parent_list_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/rocCurve/'.format(
        project_id,
        parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/frozenModels/{}/'.format(
        project_id,
        model_id
    )


@responses.activate
def test_get_validation_roc_curve(roc_curve_validation_data,
                                  roc_curve_validation_data_url,
                                  project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(roc_curve_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    roc = model.get_roc_curve('validation')

    assert roc.source == roc_curve_validation_data['source']
    assert roc.negative_class_predictions == roc_curve_validation_data['negativeClassPredictions']
    assert roc.positive_class_predictions == roc_curve_validation_data['positiveClassPredictions']
    assert roc.roc_points == from_api(roc_curve_validation_data['rocPoints'])
    assert roc.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_roc_curve_no_fallback(roc_curve_validation_data_url,
                                                     project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_validation_data_url,
        status=404
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_roc_curve('validation')


@responses.activate
@pytest.mark.usefixtures('known_warning')
def test_get_frozen_validation_roc_curve_with_fallback(
        roc_curve_parent_model_validation_data,
        roc_curve_validation_data_url,
        roc_curve_parent_model_validation_data_url,
        frozen_model_url,
        project_id, model_id, parent_model_id,
        frozen_json):
    responses.add(
        responses.GET,
        roc_curve_validation_data_url,
        status=404
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        roc_curve_parent_model_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(roc_curve_parent_model_validation_data)
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc = model.get_roc_curve('validation', fallback_to_parent_insights=True)

    assert roc.source == roc_curve_parent_model_validation_data['source']
    assert roc.negative_class_predictions == \
        roc_curve_parent_model_validation_data['negativeClassPredictions']
    assert roc.positive_class_predictions == \
        roc_curve_parent_model_validation_data['positiveClassPredictions']
    assert roc.roc_points == from_api(roc_curve_parent_model_validation_data['rocPoints'])
    assert roc.source_model_id == parent_model_id


@responses.activate
def test_get_all_roc_curves(roc_curve_validation_data,
                            roc_curve_list_url,
                            project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [roc_curve_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data['source']
    assert roc_list[0].negative_class_predictions == roc_curve_validation_data[
        'negativeClassPredictions']
    assert roc_list[0].positive_class_predictions == roc_curve_validation_data[
        'positiveClassPredictions']
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data['rocPoints'])
    assert roc_list[0].source_model_id == model_id


@responses.activate
def test_get_frozen_all_roc_curves_no_fallback(roc_curve_validation_data,
                                               roc_curve_list_url,
                                               project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [roc_curve_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data['source']
    assert roc_list[0].negative_class_predictions == roc_curve_validation_data[
        'negativeClassPredictions']
    assert roc_list[0].positive_class_predictions == roc_curve_validation_data[
        'positiveClassPredictions']
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data['rocPoints'])
    assert roc_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures('known_warning')
def test_get_forzen_all_roc_curves_with_fallback(roc_curve_validation_data,
                                                 roc_curve_parent_model_validation_data,
                                                 roc_curve_parent_model_holdout_data,
                                                 roc_curve_list_url,
                                                 roc_curve_parent_list_url,
                                                 frozen_model_url,
                                                 project_id, model_id,
                                                 parent_model_id,
                                                 frozen_json):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [roc_curve_validation_data]})
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        roc_curve_parent_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [roc_curve_parent_model_validation_data,
                                    roc_curve_parent_model_holdout_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves(fallback_to_parent_insights=True)

    assert len(roc_list) == 2
    assert roc_list[0].source == roc_curve_validation_data['source']
    assert roc_list[0].negative_class_predictions == roc_curve_validation_data[
        'negativeClassPredictions']
    assert roc_list[0].positive_class_predictions == roc_curve_validation_data[
        'positiveClassPredictions']
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data['rocPoints'])
    assert roc_list[0].source_model_id == model_id

    assert roc_list[1].source == roc_curve_parent_model_holdout_data['source']
    assert roc_list[1].negative_class_predictions == roc_curve_parent_model_holdout_data[
        'negativeClassPredictions']
    assert roc_list[1].positive_class_predictions == roc_curve_parent_model_holdout_data[
        'positiveClassPredictions']
    assert roc_list[1].roc_points == from_api(roc_curve_parent_model_holdout_data['rocPoints'])
    assert roc_list[1].source_model_id == parent_model_id


def test_get_best_f1_threshold(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When calculate recommended threshold
    best_threshold = roc.get_best_f1_threshold()
    # Then f1 score for that threshold maximal from all ROC curve points
    best_f1 = roc.estimate_threshold(best_threshold)['f1_score']
    assert all(best_f1 >= roc_point['f1_score'] for roc_point in roc.roc_points)


def test_estimate_threshold_equal(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold equal to one of precalculated
    threshold = roc_curve_validation_data['rocPoints'][1]['threshold']
    # Then estimate_threshold return valid data
    assert roc.estimate_threshold(threshold)['threshold'] == threshold


def test_estimate_threshold_new(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold from outside of precalculated
    threshold = roc_curve_validation_data['rocPoints'][1]['threshold'] + 0.1
    # Then estimate_threshold return data for next threshold bigger then requested
    assert roc.estimate_threshold(threshold)['threshold'] > threshold



import datetime
import unittest
import pandas as pd
import pytest
import responses

from datarobot import errors
from datarobot.client import Client, set_client
from datarobot.errors import DataRobotDeprecationWarning
from datarobot.utils import (dataframe_to_buffer, parse_time, get_id_from_response,
                             from_api, deprecation, recognize_sourcedata, to_api, is_urlsource)
from datarobot.utils.sourcedata import list_of_records_to_buffer
from tests.test_helpers import fixture_file_path
from tests.utils import warns


class TestDataframeSerialization(unittest.TestCase):

    def test_list_of_records_to_buffer(self):
        records = [{'a': 'b', 'c': 'd'}, {'a': 'x', 'c': 'y'}]
        expected = "a,c\r\nb,d\r\nx,y\r\n"
        buff = list_of_records_to_buffer(records)
        self.assertEqual(buff.read(), expected)

    def test_no_index_please(self):
        df = pd.DataFrame({'a': range(100), 'b': range(100)})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ['a', 'b'])

    def test_parse_time(self):
        self.assertEqual('BAD TIME', parse_time('BAD TIME'))  # returns value
        test_string_time = datetime.datetime.now().isoformat()
        self.assertIsInstance(test_string_time, str)
        self.assertIsInstance(parse_time(test_string_time), datetime.datetime)

    @responses.activate
    @pytest.mark.usefixtures('no_client_version_check')
    def test_get_id_from_response_location_header(self):
        responses.add(responses.POST,
                      'http://nothing/',
                      body='',
                      adding_headers={'Location': 'http://nothing/f-id/'})
        client = set_client(Client(token='no_matter',
                                   endpoint='http://nothing'))
        resp = client.post('')
        self.assertEqual(get_id_from_response(resp), 'f-id')

    def test_csv_buffer_is_quoted(self):
        """"
        This tests that dataframes are correctly quoted when saved back to csv, so
        that columns containing line endings won't be interpreted as multiple rows
        See: PRED-3203
        """
        df = pd.DataFrame({'A': ['test\rtest', 2], 'B': ['test\n\n\n\r\ntest', 0.75]})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        assert df.shape[0] == readback.shape[0]


class TestFromAPI(unittest.TestCase):

    def test_nested_list_of_objects_all_changed(self):
        source = {
            'oneFish': [
                {'twoFish': 'redFish'},
                {'blueFish': 'noFish'}
            ]
        }
        result = from_api(source)
        inner = result['one_fish']
        self.assertEqual(inner[0]['two_fish'], 'redFish')
        self.assertEqual(inner[1]['blue_fish'], 'noFish')

    def test_nested_objects_all_changed(self):
        source = {
            'oneFish': {
                'twoFish': 'redFish'
            }
        }

        result = from_api(source)
        self.assertEqual(result['one_fish']['two_fish'], 'redFish')


def test_from_api_filter_nones():
    data_in = {
        'someValue': {
            'nestedNone': None,
            'nestedNotNone': ''
        },
        'listNone': [{
            'someValue': None,
            'someNotNone': ''
        }],
        'otherNone': None
    }
    data_out = {
        'some_value': {'nested_not_none': ''},
        'list_none': [{'some_not_none': ''}]
    }
    assert data_out == from_api(data_in)


def test_from_api_no_filter_nones():
    data_in = {
        'someValue': {
            'nestedNone': None,
            'nestedNotNone': ''
        },
        'listNone': [{
            'someValue': None,
            'someNotNone': ''
        }],
        'otherNone': None
    }
    data_out = {
        'some_value': {
            'nested_none': None,
            'nested_not_none': ''
        },
        'list_none': [{
            'some_value': None,
            'some_not_none': ''
        }],
        'other_none': None
    }
    assert data_out == from_api(data_in, keep_null_keys=True)


def test_from_api_keep_nones():
    data_in = {
        'someValue': {
            'nestedNone': None,
            'nestedNotNone': '',
            'levelTwo': {
                'levelThree': None
            }
        },
        'otherNone': None,
        'keepLevelOne': None
    }
    data_out = {
        'some_value': {
            'nested_not_none': '',
            'nested_none': None,
            'level_two': {
                'level_three': None
            }
        },
        'keep_level_one': None
    }
    assert data_out == from_api(data_in, keep_attrs=[
        'some_value.nested_none',
        'some_value.level_two.level_three',
        'keep_level_one'
    ])


def test_from_api_multiple_underscore_ignored():
    in_data = {'samplePct__gte': 64}
    expected = {'sample_pct__gte': 64}
    assert from_api(in_data) == expected


def test_to_api_multiple_underscore_ignored():
    in_data = {'sample_pct__gte': 64}
    expected = {'samplePct__gte': 64}
    assert to_api(in_data) == expected


def test_from_api_single_letter():
    in_data = {'iAmAKey': 'test'}
    expected = {'i_am_a_key': 'test'}
    assert from_api(in_data) == expected


def test_to_api_single_letter():
    in_data = {'i_am_a_key': 'test'}
    expected = {'iAmAKey': 'test'}
    assert to_api(in_data) == expected


@deprecation.deprecated(deprecated_since_version='v0.1.2',
                        will_remove_version='v1.2.3')
def bar(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        arg1 + arg2

    """
    return arg1 + arg2


@deprecation.deprecated(deprecated_since_version='v0.1.2',
                        will_remove_version='v1.2.3',
                        message='Use of `bar` is recommended instead.')
def foo(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        foo + bar

    """
    return arg1 + arg2


def test_deprecation_with_message(known_warning):
    foo(1, 2)
    w_msg = known_warning[0].message.args[0]
    assert w_msg == \
        '`foo` has been deprecated in `v0.1.2`, will be removed ' \
        'in `v1.2.3`. Use of `bar` is recommended instead.'


def test_deprecation_message(known_warning):
    bar(1, 2)
    w_instance = known_warning[0].message
    assert isinstance(w_instance, DataRobotDeprecationWarning)


class TestSourcedataUtils(unittest.TestCase):
    def setUp(self):
        self.default_fname = 'predict.csv'

    def test_recognize_sourcedata_passed_dataframe(self):
        df = pd.DataFrame({'a': range(100), 'b': range(100)})
        kwargs = recognize_sourcedata(df, self.default_fname)
        self.assertTrue('filelike' in kwargs)
        self.assertEqual(kwargs.get('fname'), self.default_fname)
        buff = kwargs['filelike']
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ['a', 'b'])

    def test_recognize_sourcedata_passed_filelike(self):
        path = fixture_file_path('synthetic-100.csv')
        with open(path, 'rb') as fd:
            kwargs = recognize_sourcedata(fd, self.default_fname)
            self.assertTrue(kwargs.get('filelike') is fd)
            self.assertEqual(kwargs.get('fname'), self.default_fname)

    def test_recognize_sourcedata_passed_filepath(self):
        file_path = fixture_file_path('synthetic-100.csv')
        kwargs = recognize_sourcedata(file_path, self.default_fname)
        self.assertEqual(kwargs.get('file_path'), file_path)
        self.assertEqual(kwargs.get('fname'), 'synthetic-100.csv')

    def test_recognize_sourcedata_passed_content(self):
        content = b'a,b,c\n' + b'1,2,3\n' * 100
        kwargs = recognize_sourcedata(content, self.default_fname)
        self.assertEqual(kwargs.get('content'), content)
        self.assertEqual(kwargs.get('fname'), self.default_fname)

    def test_is_urlsource_passed_true(self):
        result = is_urlsource('http://path_to_urlsource')
        self.assertTrue(result)

    def test_is_urlsource_file_passed_true(self):
        result = is_urlsource('file://path_to_urlsource')
        self.assertTrue(result)

    def test_is_urlsource_file_passed_false(self):
        result = is_urlsource('filename')
        self.assertFalse(result)

    def test_is_urlsource_passed_false(self):
        result = is_urlsource('not_a_path_to_urlsource')
        self.assertFalse(result)

    def test_is_urlsource_s3_passed_true(self):
        result = is_urlsource('s3://path/to/a/bucket/data.csv')
        self.assertTrue(result)

    def test_fatfingered_filepath_raises(self):
        content = b'/home/datarobot/mistypefilepath.csv'
        with self.assertRaises(errors.InputNotUnderstoodError):
            recognize_sourcedata(content, self.default_fname)


def test_resource():
    """Python 3 ResourceWarning of implicitly closed files."""
    with warns():
        open(__file__)


class TestImports(unittest.TestCase):
    def test_rename(self):
        """datarobot_sdk rename."""
        with self.assertRaises(ImportError):
            import datarobot_sdk  # noqa



import json

import pytest
import responses

import datarobot as dr
from datarobot.utils import camelize
from tests.utils import request_body_to_json


class TestHDFSCreate(object):
    @pytest.fixture
    def hdfs_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = '{}/{}/'.format(unittest_endpoint, 'hdfsProjects')
        responses.add(
            responses.POST,
            mysql_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @responses.activate
    @pytest.mark.usefixtures('project_creation_responses',
                             'hdfs_create_response')
    def test_create_hdfs_minimum(self):
        dr.Project.create_from_hdfs(url='hdfs://webhdfs.com/directory/file.csv')
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body['url'] == 'hdfs://webhdfs.com/directory/file.csv'

    @responses.activate
    @pytest.mark.usefixtures('project_creation_responses',
                             'hdfs_create_response')
    def test_create_hdfs_all_options(self):
        dr.Project.create_from_hdfs(url='hdfs://webhdfs.com/directory/file.csv',
                                    port=9999,
                                    project_name='HDFS Project',
                                    max_wait=100)
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body['url'] == 'hdfs://webhdfs.com/directory/file.csv'
        assert create_body['port'] == 9999
        assert create_body['projectName'] == 'HDFS Project'


class TestDataSourceCreate(object):
    @pytest.fixture
    def data_source_create_response(self, unittest_endpoint, async_url):
        project_create_url = '{}/{}'.format(unittest_endpoint, dr.Project._path)
        responses.add(
            responses.POST,
            project_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @pytest.fixture
    def data_source(self):
        return dr.DataSource('5acc8437ec8d670001ba16bf')

    @responses.activate
    @pytest.mark.usefixtures('project_creation_responses',
                             'data_source_create_response')
    def test_create_from_data_source_id(self):
        username = 'test_user'
        password = 'megasuperubersecretpassword'
        data_source_id = '5acc8437ec8d670001ba16bf'
        project_name = 'data source project'
        dr.Project.create_from_data_source(data_source_id, username, password,
                                           project_name=project_name)
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {
            'user': username,
            'password': password,
            'dataSourceId': data_source_id,
            'projectName': project_name
        }


class TestDatasetCreate(object):
    @responses.activate
    @pytest.mark.usefixtures('project_creation_responses')
    def test_create_from_dataset_use_defaults(self, project_collection_url,
                                              project_without_target_json):
        dataset_id = 'abc123'
        project = dr.Project.create_from_dataset(dataset_id)

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {'datasetId': dataset_id}
        assert responses.calls[0].request.url == project_collection_url

        response_json = json.loads(project_without_target_json)
        assert project.id == response_json['id']
        assert project.project_name == response_json['projectName']
        assert project.file_name == response_json['fileName']
        assert project.stage == response_json['stage']

    @responses.activate
    @pytest.mark.usefixtures('project_creation_responses')
    @pytest.mark.parametrize('param_name,param_value', [
        ('dataset_version_id', '20gjkfdlj'),
        ('project_name', 'my cool name'),
        ('user', 'that guy over there'),
        ('password', 'my voice is my passport'),
        ('credential_id', '12345'),
        ('use_kerberos', True)
    ])
    def test_create_from_dataset_using_params(self, param_name, param_value):
        dataset_id = 'abc123'
        dr.Project.create_from_dataset(dataset_id, **{param_name: param_value})
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {'datasetId': dataset_id, camelize(param_name): param_value}



import json
import os
import uuid
from collections import defaultdict
from copy import deepcopy

import responses
from requests_toolbelt import MultipartEncoder

from datarobot import CustomModelVersion

import pytest

from tests.model.utils import assert_custom_model_version


@pytest.fixture
def mocked_version(mocked_custom_model_version):
    return mocked_custom_model_version


@pytest.fixture
def make_versions_url(unittest_endpoint):
    def _make_versions_url(environment_id, version_id=None):
        base_url = "{}/customModels/{}/versions/".format(
            unittest_endpoint, environment_id
        )
        if version_id is not None:
            return "{}{}/".format(base_url, version_id)
        return base_url

    return _make_versions_url


def mock_get_response(url, response):
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(response),
    )


def test_from_server_data(mocked_version):
    version = CustomModelVersion.from_server_data(mocked_version)
    assert_custom_model_version(version, mocked_version)


@responses.activate
def test_get_version(mocked_version, make_versions_url):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    # act
    version = CustomModelVersion.get(
        mocked_version["customModelId"], mocked_version["id"]
    )

    # assert
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions(mocked_versions, make_versions_url):
    # arrange
    custom_model_id = mocked_versions["data"][0]["customModelId"]
    url = make_versions_url(custom_model_id)
    mock_get_response(url, mocked_versions)

    # act
    versions = CustomModelVersion.list(custom_model_id)

    # assert
    assert len(versions) == len(mocked_versions["data"])
    for version, mocked_version in zip(versions, mocked_versions["data"]):
        assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions_multiple_pages(mocked_versions, make_versions_url):
    # arrange
    custom_model_id = mocked_versions["data"][0]["customModelId"]

    url1 = make_versions_url(custom_model_id)
    url2 = make_versions_url(custom_model_id) + "2"

    mocked_versions_2nd = deepcopy(mocked_versions)
    mocked_versions["next"] = url2

    mock_get_response(url1, mocked_versions)
    mock_get_response(url2, mocked_versions_2nd)

    # act
    versions = CustomModelVersion.list(custom_model_id)

    # assert
    assert len(versions) == len(mocked_versions["data"]) + len(
        mocked_versions_2nd["data"]
    )
    for version, mocked_version in zip(
        versions, mocked_versions["data"] + mocked_versions_2nd["data"]
    ):
        assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url1
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url2


@responses.activate
def test_download(mocked_version, make_versions_url, tmpdir):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)
    responses.add(
        responses.GET,
        url + "download/",
        status=200,
        content_type="application/json",
        body=b"content",
    )

    downloaded_file = tmpdir.mkdir("sub").join("download")
    downloaded_file_path = str(downloaded_file)

    # act
    version = CustomModelVersion.get(
        mocked_version["customModelId"], mocked_version["id"]
    )
    version.download(downloaded_file_path)

    # assert
    downloaded_file.read() == b"content"

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "download/"


@responses.activate
def test_update(mocked_version, make_versions_url):
    # arrange
    attrs = {"description": "xx"}

    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    mocked_version.update(attrs)

    responses.add(
        responses.PATCH,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    version.update(**attrs)

    # assert
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "PATCH"
    assert responses.calls[1].request.url == url
    assert responses.calls[1].request.body == json.dumps(attrs).encode()


@responses.activate
def test_refresh(mocked_version, make_versions_url):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    mocked_version.update({"description": "xx"})

    mock_get_response(url, mocked_version)

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    version.refresh()

    # assert
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url


@responses.activate
def test_create_clean(mocked_version, make_versions_url, tmpdir):
    url = make_versions_url(mocked_version["customModelId"])

    responses.add(
        responses.POST,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    file = tmpdir.mkdir("sub").join("file")
    file.write(b"content")
    file_path = str(file)

    version = CustomModelVersion.create_clean(
        mocked_version["customModelId"],
        True,
        files=[(file_path, "/d/file.txt")]
    )
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url


@responses.activate
def test_create_from_previous(mocked_version, make_versions_url, tmpdir):
    url = make_versions_url(mocked_version["customModelId"])

    responses.add(
        responses.PATCH,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    file = tmpdir.mkdir("sub").join("file")
    file.write(b"content")
    file_path = str(file)

    version = CustomModelVersion.create_from_previous(
        mocked_version["customModelId"],
        True,
        files=[(file_path, "/d/file.txt")],
        files_to_delete=["2323423423423423"]
    )
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "PATCH"
    assert responses.calls[0].request.url == url


@responses.activate
@pytest.mark.parametrize('method', [
    (CustomModelVersion.create_clean, "POST"),
    (CustomModelVersion.create_from_previous, "PATCH"),
])
@pytest.mark.parametrize('expected_folder_paths', [
    None,
    [],
    ['a.zip', 'a/a.zip', 'sub/a.txt', 'sub/b.pdf', 'sub/sub2/c.rtf'],
])
@pytest.mark.parametrize('expected_file_paths', [
    None,
    [],
    ['y.xyz', 'dd/ttt.txt']
])
def test_create_folder_and_files(
    mocked_version,
    make_versions_url,
    tmpdir,
    method,
    expected_folder_paths,
    expected_file_paths
):
    custom_model_version_method = method[0]
    http_method = method[1]

    # arrange
    responses.add(
        http_method,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        body=json.dumps(mocked_version)
    )

    # arrange: create folder with required structure to be uploaded
    folder_path = None
    if expected_folder_paths:
        folder_path = tmpdir.mkdir(str(uuid.uuid4()))
        for expected_path in expected_folder_paths:
            path_items, file = os.path.split(expected_path)
            d = folder_path
            if path_items:
                for path_item in path_items.split('/'):
                    if os.path.exists(os.path.join(str(d), path_item)):
                        d = d.join(path_item)
                    else:
                        d = d.mkdir(path_item)
            d.join(file).write(b"content")
        folder_path = str(folder_path)

    # arrange: create files to be uploaded
    files = None
    if expected_file_paths:
        files = []
        for expected_path in expected_file_paths:
            file = tmpdir.join(str(uuid.uuid4()))
            file.write(b"content")
            files.append((str(file), expected_path))

    # act
    custom_model_version_method(
        mocked_version["customModelId"],
        True,
        folder_path=folder_path,
        files=files,
    )

    # assert
    req = responses.calls[0].request
    assert req.method == http_method
    assert req.url == make_versions_url(mocked_version["customModelId"])
    assert isinstance(req.body, MultipartEncoder)

    # get fields submitted with request
    fields = defaultdict(list)
    for name, value in req.body.fields:
        fields[name].append(value)

    assert len(fields["isMajorUpdate"]) == 1
    assert fields["isMajorUpdate"][0] == 'True'

    files = fields["file"]
    file_paths = fields["filePath"]

    # verify that files & filePaths make pairs
    assert len(files) == len(file_paths)
    for file, file_path in zip(files, file_paths):
        assert file[0] == os.path.basename(file_path)

    expected_folder_paths = expected_folder_paths or []
    expected_file_paths = expected_file_paths or []

    # There's an option to create a custom model version by passing a pass to a directory.
    # For such a directory, files are created with expected_folder_paths.
    # As there's no control in what order filesystem reads files from a such a directory,
    # convert paths to set and then assert
    assert set(file_paths) == set(expected_folder_paths) | set(expected_file_paths)



import pytest
import responses
from mock import patch

from datarobot.errors import ClientError
from datarobot.models import ExternalResidualsChart


class TestExternalResiduals(object):

    @pytest.fixture
    def url(self, project_id, model_id):
        return ('https://host_name.com/'
                'projects/{}/models/{}/datasetResidualsCharts/').format(project_id, model_id)

    @pytest.fixture
    def residuals_response(self):
        return [
            {
                'datasetId': 'datasetId{}'.format(i),
                'residualMean': 1.234567,
                'coefficientOfDetermination': 0.65,
                'standardDeviation': 0.1,
                'data': [
                    [1, 2.5, 1.5, 16],
                    [3, 4.6, 1.6, None]
                 ],
                'histogram': [
                     {'intervalStart': 2, 'intervalEnd': 4, 'occurrences': 1},
                     {'intervalStart': 4, 'intervalEnd': 6, 'occurrences': 1}
                ]
            } for i in range(20)]

    def test_external_residuals_parsing(self, residuals_response):
        residuals = ExternalResidualsChart.from_server_data(residuals_response[0])
        assert residuals.dataset_id == residuals_response[0]['datasetId']
        assert residuals.residual_mean == residuals_response[0]['residualMean']
        cod = residuals_response[0]['coefficientOfDetermination']
        assert residuals.coefficient_of_determination == cod
        assert residuals.standard_deviation == residuals_response[0]['standardDeviation']
        assert residuals.data == [tuple(x) for x in residuals_response[0]['data']]

    @responses.activate
    def test_get_residuals_ok(self, url, project_id, model_id, residuals_response):
        dataset_id = 'datasetId0'
        url_page = url + '?offset=0&limit=1&datasetId={}'.format(dataset_id)
        resp_data = dict(count=1, previous=None, next=None, data=[residuals_response[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        chart = ExternalResidualsChart.get(project_id, model_id, dataset_id)
        assert chart.dataset_id == residuals_response[0]['datasetId']
        assert chart.residual_mean == residuals_response[0]['residualMean']
        cod = residuals_response[0]['coefficientOfDetermination']
        assert chart.coefficient_of_determination == cod
        assert chart.standard_deviation == residuals_response[0]['standardDeviation']
        assert chart.data == [tuple(x) for x in residuals_response[0]['data']]

    @responses.activate
    def test_get_residuals_failure(self, url, project_id, model_id):
        dataset_id = 'dataset_id_0'
        url_page = url + '?offset=0&limit=1&datasetId={}'.format(dataset_id)
        resp_data = dict(count=0, previous=None, next=None, data=[])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        with pytest.raises(ClientError, match='Requested residual chart does not exist.'):
            ExternalResidualsChart.get(project_id, model_id, dataset_id)

    @responses.activate
    @patch(
        'datarobot.models.external_dataset_scores_insights.'
        'external_dataset_residuals.DEFAULT_BATCH_SIZE',
        10
    )
    def test_residuals_all_results(self, url, residuals_response, project_id, model_id):
        url_page1 = url + '?limit=10&offset=0'
        url_page2 = url + '?limit=10&offset=10'
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=residuals_response[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=residuals_response[10:])
        responses.add(responses.GET, url_page1, json=resp_data_1, status=200,
                      match_querystring=True)
        responses.add(responses.GET, url_page2, json=resp_data_2, status=200,
                      match_querystring=True)

        residuals = ExternalResidualsChart.list(project_id, model_id, limit=0)
        assert len(residuals) == 20
        for i, chart in enumerate(residuals):
            assert isinstance(chart, ExternalResidualsChart)
            assert chart.dataset_id == 'datasetId{}'.format(i)
            assert chart.data == [tuple(x) for x in residuals_response[0]['data']]



import json
from collections import OrderedDict

import dateutil
import pytest
import responses

from datarobot.models import Deployment, ServiceStats, ServiceStatsOverTime
from datarobot.utils import from_api


class TestServiceStats(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            'model_id': deployment_data['model']['id'],
            'period': {
                'start': '2019-08-01T00:00:00.000000Z',
                'end': '2019-08-10T00:00:00.000000Z'
            },
            'metrics': {
                'totalPredictions': 1000,
                'totalRequests': 50,
            }
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = '{}/deployments/{}/serviceStats/'.format(unittest_endpoint, deployment_data['id'])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(response_data)
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    def test_service_stats(self, deployment_data, response_data):
        service_stats_objs = [
            ServiceStats.get(deployment_data['id']),
            Deployment.get(deployment_data['id']).get_service_stats()
        ]
        for service_stats in service_stats_objs:
            assert service_stats.model_id == response_data['model_id']

            expected_start_date = dateutil.parser.parse(response_data['period']['start'])
            expected_end_date = dateutil.parser.parse(response_data['period']['end'])
            assert service_stats.period['start'] == expected_start_date
            assert service_stats.period['end'] == expected_end_date

            assert service_stats.metrics.keys() == response_data['metrics'].keys()
            for metric_name, metric in service_stats.metrics.items():
                expected_value = response_data['metrics'][metric_name]
                assert metric == expected_value
                assert service_stats[metric_name] == expected_value

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    @pytest.mark.parametrize('kwarg, expected_query_string', [
        ({'model_id': 'abc'}, '?modelId=abc'),
        ({'execution_time_quantile': 0.74}, '?executionTimeQuantile=0.74'),
        ({'response_time_quantile': 0.43}, '?responseTimeQuantile=0.43'),
        ({'slow_requests_threshold': 40}, '?slowRequestsThreshold=40'),
    ])
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStats.get(deployment_data['id'], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data['id']).get_service_stats(**kwarg)
        assert expected_query_string in responses.calls[0].request.url


class TestServiceStatsOverTime(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            'model_id': deployment_data['model']['id'],
            'metric': 'totalPredictions',
            'summary': {
                'period': {
                    'start': '2019-08-01T00:00:00.000000Z',
                    'end': '2019-08-04T00:00:00.000000Z'
                },
                'value': 2250,
            },
            'buckets': [
                {
                    'period': {
                        'start': '2019-08-01T00:00:00.000000Z',
                        'end': '2019-08-02T00:00:00.000000Z'
                    },
                    'value': 1250,
                }, {
                    'period': {
                        'start': '2019-08-02T00:00:00.000000Z',
                        'end': '2019-08-03T00:00:00.000000Z'
                    },
                    'value': 1000,
                }, {
                    'period': {
                        'start': '2019-08-03T00:00:00.000000Z',
                        'end': '2019-08-04T00:00:00.000000Z'
                    },
                    'value': 0,
                }
            ]
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        deployment_id = deployment_data['id']
        url = '{}/deployments/{}/serviceStatsOverTime/'.format(unittest_endpoint, deployment_id)
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(response_data)
        )

    def assert_bucket(self, actual, expected):
        assert actual.keys() == expected.keys()

        actual_period = actual['period']
        if actual_period:
            assert actual['period'].keys() == expected['period'].keys()
            assert actual_period['start'] == dateutil.parser.parse(expected['period']['start'])
            assert actual_period['end'] == dateutil.parser.parse(expected['period']['end'])
        else:
            assert actual_period is None

        assert actual['value'] == expected['value']

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    def test_service_stats_over_time(self, deployment_data, response_data):
        service_stats_over_time_objs = [
            ServiceStatsOverTime.get(deployment_data['id']),
            Deployment.get(deployment_data['id']).get_service_stats_over_time()
        ]
        for service_stats_over_time in service_stats_over_time_objs:
            expected = from_api(response_data, keep_null_keys=True)
            assert service_stats_over_time.metric == expected['metric']
            assert service_stats_over_time.model_id == expected['model_id']

            self.assert_bucket(service_stats_over_time.summary, expected['summary'])
            for index, bucket in enumerate(service_stats_over_time.buckets):
                self.assert_bucket(bucket, expected['buckets'][index])

            expected_bucket_values = [
                (dateutil.parser.parse(bucket['period']['start']), bucket['value'])
                for bucket in expected['buckets']]
            assert service_stats_over_time.bucket_values == OrderedDict(expected_bucket_values)

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    @pytest.mark.parametrize('kwarg, expected_query_string', [
        ({'metric': 'LogLoss'}, '?metric=LogLoss'),
        ({'model_id': 'abc'}, '?modelId=abc'),
        ({'bucket_size': 'P1D'}, '?bucketSize=P1D'),
        ({'quantile': 0.74}, '?quantile=0.74'),
        ({'threshold': 0.43}, '?threshold=0.43'),
    ])
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStatsOverTime.get(deployment_data['id'], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data['id']).get_service_stats_over_time(**kwarg)
        assert expected_query_string in responses.calls[0].request.url



from copy import deepcopy
import json
import pytest

import responses

from datarobot import (
    Job, TrainingPredictionsJob, FeatureImpactJob, Project, enums, TrainingPredictions)
from datarobot.errors import AsyncTimeoutError, AsyncProcessUnsuccessfulError

mock_job_data = {
    'status': 'inprogress',
    'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
    'id': '1',
    'jobType': 'model',
    'isBlocked': False,
    'modelId': '5d518cd3962d741512605e2b',
    'projectId': 'p-id'
}

mock_model_job_data = {
    'status': enums.QUEUE_STATUS.INPROGRESS,
    'processes': [
        'One-Hot Encoding',
        'Bernoulli Naive Bayes classifier (scikit-learn)',
        'Missing Values Imputed',
        'Gaussian Naive Bayes classifier (scikit-learn)',
        'Naive Bayes combiner classifier',
        'Calibrate predictions'
    ],
    'projectId': 'p-id',
    'samplePct': 28.349,
    'modelType': 'Naive Bayes combiner classifier',
    'featurelistId': '56d8620bccf94e26f37af0a3',
    'modelCategory': 'model',
    'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
    'id': '1'
}

SAMPLE_BINARY_PREDICTION_RETURN = json.dumps({
    u'positiveClass': 1.0,
    u'task': u'Binary',
    u'predictions': [
        {u'positiveProbability': 0.9, u'prediction': 1.0,
         u'rowId': 0, u'predictionValues': [{u'value': 0.9, u'label': u'1.0'},
                                            {u'value': 0.1, u'label': u'0.0'}]},
        {u'positiveProbability': 0.1, u'prediction': 0.0,
         u'rowId': 1, u'predictionValues': [{u'value': 0.9, u'label': u'1.0'},
                                            {u'value': 0.1, u'label': u'0.0'}]},
        {u'positiveProbability': 0.8, u'prediction': 1.0,
         u'rowId': 2, u'predictionValues': [{u'value': 0.8, u'label': u'1.0'},
                                            {u'value': 0.2, u'label': u'0.0'}]}
    ]
})


def assert_job_attributes(job):
    assert job.id == 1
    assert job.project == Project('p-id')
    assert job.status == 'inprogress'
    assert job.job_type == 'model'
    assert job.model_id == '5d518cd3962d741512605e2b'
    # ensure the instantiated object has the model ID as well
    assert job.model.id == '5d518cd3962d741512605e2b'
    assert not job.is_blocked
    assert job._job_details_path == 'projects/p-id/modelJobs/1/'


def test_instantiate():
    new_job = Job(mock_job_data)
    assert_job_attributes(new_job)


def test_future_proof():
    Job(dict(mock_job_data, future='new'))


@responses.activate
@pytest.mark.parametrize('job_class', (Job, FeatureImpactJob))
def test_get_completed(client, job_class):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_job_data, status='COMPLETED')),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    job = job_class.get('p-id', '1')
    assert job.status == 'COMPLETED'
    assert job._completed_resource_url == resource_url


@responses.activate
def test_get_completed_feature_impact(client, feature_impact_server_data):
    job_result_url = 'https://host_name.com/resource-location/'
    pid = 'p-id'

    # Get project.
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/'.format(pid),
        status=200,
        body=json.dumps(
            {
                'id': pid,
            }
        ),
        content_type='application/json',
    )
    project = Project.get(pid)

    # Get not finished jobs.
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/jobs/'.format(pid),
        status=200,
        body=json.dumps(
            {
                'jobs': [mock_job_data],
            }
        ),
        content_type='application/json',
    )
    job = project.get_all_jobs()[0]

    # Mark job as finished.
    job_data = deepcopy(mock_job_data)
    job_data['status'] = 'COMPLETED'
    job_data['jobType'] = 'featureImpact'

    # Get that finished job and its results.
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/jobs/1/'.format(pid),
        status=303,
        body=json.dumps(job_data),
        content_type='application/json',
        adding_headers={'Location': job_result_url},
    )
    responses.add(
        responses.GET,
        job_result_url,
        status=200,
        body=json.dumps(feature_impact_server_data),
        content_type='application/json',
    )
    assert job.get_result() == feature_impact_server_data['featureImpacts']


@responses.activate
@pytest.mark.parametrize('job_class', (Job, FeatureImpactJob))
def test_get_no_model_id(client, job_class):
    resource_url = 'https://host_name.com/resource-location/'
    job_data = mock_job_data.copy()
    del job_data['modelId']
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(job_data),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    job = job_class.get('p-id', '1')
    assert job.model_id is None
    assert job.model is None


@responses.activate
def test_get_completed_pred(client):
    get_predictions_url = 'https://host_name.com/projects/p-id/predictions/1/'

    mock_generic_job_data = {
        'status': 'COMPLETED',
        'url': 'https://host_name.com/projects/p-id/predictJobs/1/',
        'id': '1',
        'jobType': 'predict',
        'isBlocked': False,
        'projectId': 'p-id'
    }

    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_generic_job_data, status='COMPLETED')),
        content_type='application/json',
        adding_headers={'Location': get_predictions_url}
    )
    responses.add(
        responses.GET,
        get_predictions_url,
        status=200,
        body=SAMPLE_BINARY_PREDICTION_RETURN,
        content_type='application/json'
    )

    job = Job.get('p-id', '1')

    data = job.get_result()
    assert job.status == 'COMPLETED'

    # inspect the data
    expected_columns = ['positive_probability', 'prediction', 'row_id', 'class_0.0', 'class_1.0']
    assert sorted(data.columns.values) == sorted(expected_columns)


@responses.activate
def test_wait_for_completion_never_finishes(client, mock_async_time):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(mock_job_data),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    job = Job(mock_job_data)
    mock_async_time.time.side_effect = (0, 2)
    with pytest.raises(AsyncTimeoutError):
        job.wait_for_completion(max_wait=1)


@responses.activate
@pytest.mark.parametrize('job_class', (Job, FeatureImpactJob))
def test_wait_for_completion_finishes(client, mock_async_time, job_class):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_job_data, status='COMPLETED')),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    mock_async_time.time.side_effect = (0, 2)
    job = job_class(mock_job_data)
    job.wait_for_completion(max_wait=3)
    assert job.status == 'COMPLETED'


@responses.activate
def test_get_result_errored(project_id, project_url):
    job_url = '{}jobs/1/'.format(project_url)
    mock_job = {
        'status': enums.QUEUE_STATUS.ERROR,
        'url': '{}modelJobs/1/'.format(project_url),
        'id': '1',
        'jobType': 'model',
        'isBlocked': False,
        'projectId': project_id
    }
    responses.add(responses.GET, job_url, status=200,
                  json=mock_job, content_type='application/json')
    job = Job.get(project_id, '1')
    with pytest.raises(AsyncProcessUnsuccessfulError):
        job.get_result()


@responses.activate
@pytest.mark.parametrize('job_class', (Job, FeatureImpactJob))
def test_get(client, job_class):
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(mock_job_data),
        content_type='application/json'
    )
    job = job_class.get('p-id', '1')
    assert_job_attributes(job)
    assert job._completed_resource_url is None


@responses.activate
def test_refresh(client):
    job = Job(mock_job_data)
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(refresh_job_data),
        content_type='application/json'
    )
    job.refresh()
    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_cancel_job(client):
    responses.add(
        responses.DELETE,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=204,
        body='',
        content_type='application/json'
    )
    generic_job = Job(mock_job_data)
    generic_job.cancel()


@responses.activate
def test_training_predictions_job__refresh__ok():
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.ERROR)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(refresh_job_data),
        content_type='application/json'
    )

    job = TrainingPredictionsJob(mock_job_data, model_id='model-id', data_subset='all')
    job.refresh()

    assert job.status == enums.QUEUE_STATUS.ERROR


@responses.activate
def test_training_predictions_job__get__ok():
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(mock_job_data),
        content_type='application/json'
    )

    job = TrainingPredictionsJob.get('p-id', '1', model_id='model-id', data_subset='all')

    assert job.status == enums.QUEUE_STATUS.INPROGRESS


@responses.activate
def test_training_predictions_job__wait_for_completion__ok():
    resource_url = 'https://host_name.com/resource-location/'
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(refresh_job_data),
        content_type='application/json',
        adding_headers={'Location': resource_url},
    )

    job = TrainingPredictionsJob(mock_job_data, model_id='model-id', data_subset='all')
    job.wait_for_completion(max_wait=3)

    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_regression_test_get_completed_pred_with_training_predictions_job(client):

    project_id = 'p-id'
    prediction_id = '1'
    data_path = "projects/{project_id}/trainingPredictions/{prediction_id}".format(
        project_id=project_id, prediction_id=prediction_id)
    get_predictions_url = 'https://host_name.com/{data_path}/'.format(data_path=data_path)
    mock_generic_job_data = {
        'status': enums.QUEUE_STATUS.COMPLETED,
        'url': 'https://host_name.com/projects/p-id/predictJobs/1/',
        'id': prediction_id,
        'jobType': 'a job',
        'isBlocked': False,
        'projectId': project_id
    }

    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_generic_job_data)),
        content_type='application/json',
        adding_headers={'Location': get_predictions_url}
    )

    job = TrainingPredictionsJob.get(project_id, prediction_id)

    data = job.get_result()
    assert job.status == 'COMPLETED'

    assert isinstance(data, TrainingPredictions)
    assert data.project_id == project_id
    assert data.prediction_id == prediction_id
    assert data.path == data_path + '/'
